/*-------------------------------------------------------------------------------------------------\
| IMU 1b Microcontroller Firmware
|
| File:					IMUC.C			Mainline Code
|
| Global Routines:	Main()
|
| Local Routines:		
|
\-------------------------------------------------------------------------------------------------*/

#include	<pic.h>
#include "imuc.h"

//-------------------------------------------------------------------------------------------------
//  Define Global Variables (FOR ALL ROUTINES AND MODULES)
//-------------------------------------------------------------------------------------------------

	// Note: In PICC, "char" is an byte which is by default unsigned, "unsigned int" is a 16bit quantity.
	//
	// ALL VARIABLES ARE INITIALIZED TO ZERO ON RESET (courtesy Hi-Tech PIC C startup code)
	// Nonzero initializations (marked with "*") are initialized in the function Initialize() called in main.
	
	bank1	bit				FlgPowerUp;
	bank1	bit				FlgSampleDone;
	bank1	bit				FlgRecvdCmd;
	bank1	bit				FlgSendSamples;
	
	bank1	char				SystemError;						// stores the system error

	bank1	char				SampleMask;
	bank1	char				MasterSampleMask;
	bank1	char				LPFMask;
	bank1	char				MasterLPFMask;
	bank1 char				SendMask;

	bank1	unsigned	int	CRC;

	bank1	char				TimeStampL;							// Low byte of the TimeStamp (Snapshot of Timer2)
	bank1	unsigned long	TimeStampH;							// High 32 bits of the TimeStamp
	bank1	unsigned long	TimeCnt;								// 32 bit counter in Timer2 interrupt.

	bank1	char				SendSamplesCnt;					// The count of how many Timer2 ticks to wait before sending data

	bank1	unsigned	int	Sample[SAMPLELENGTH];

	bank1	unsigned long	Y[SAMPLELENGTH];					// Outputs of the low pass filter
	
	bank1	char				SerialQOut; 						// Index for the next message to be read out of the message queue
	bank1	char				SerialQIn;							// Index for the next message to load into the message queue
	
	bank2 char				SerialQueue[SERIALQLENGTH];	// Queue for outgoing serial bus messages to be sent to the UART

	bank3	char				RxQPtr;								// Pointer in to the Receive Command
	bank3	char				RxQueue[RXQLENGTH];				// The queue for received command bytes



	
//-------------------------------------------------------------------------------------------------	
//  Define Local (to this file) Global Variables
//-------------------------------------------------------------------------------------------------

	// None
	
//-------------------------------------------------------------------------------------------------	
//  Local #defines
//-------------------------------------------------------------------------------------------------

	// None

//-------------------------------------------------------------------------------------------------	
//  Local Function Prototypes
//-------------------------------------------------------------------------------------------------

	void	LPFSamples	(void);
	void	SendSamples	(void);
	void	ExecRcvdCmd	(void);
	void	LoadSerialQueue	(char);
	
/*-------------------------------------------------------------------------------------------------\
| Function:		main
|
| Purpose:		Run all Mainline code
|
| Discussion:	Main calls the initialization code and then goes into an infinite loop of calling
|					all mainline functions.
|
| Expects:		Interrupts to be enabled.
|					
|   Date   | Notes
|----------|---------------------------------------------------------------------------------------
| 06/09/01 | Created from other projects.
\--------------------------------------------------------------------------------------------------*/

void main (void)
{

	// Initialize the PIC & Software
	Initialize();

	// Turn on the "Mainline" LED to show that we're in the mainline.
	PORTC |= (1<<5);

	while (TRUE)
	{
		// Low Pass Filter sampled data if available

		if (LPFMask)
			LPFSamples();
			
		// Process command if one is received
		
		if (FlgRecvdCmd && !LPFMask)
			ExecRcvdCmd();

		// Pack the samples (LPF'ed or not) to the UART. Don't worry about syncing
		// with the rest of the PIC - grabbing across samples is ok, and
		// don't do it until after the LPF routine is done.
			
		if (FlgSendSamples && FlgSampleDone && !LPFMask)
			SendSamples();

		// Finally, turn on a debug LED depending on the error condition

		if (SystemError)
			PORTE |= (1<<2);
	}
}




/*-------------------------------------------------------------------------------------------------\
| Function:		LPFSamples
|
| Purpose:		Low Pass Filter any samples that are ready to be LPF'ed.
|
| Discussion:	As selected in LPFMask, the data in sample is low pass filtered. Note that SampleADC()
|					in IMUC_INT.C uses the LPFMask as a sort of array of 7 flag dictating which samples
|					are ready to be LPF'ed.
|
|					We're using an approximation of a one-pole low pass filter - namely,
|					a running average. We use the form:
|					
|						y(i) = a*y(i-1) + (1-a)*x(i)
|					
|						a = alpha = exp(-dT/Tau)      where dT = sample period [s] & tau = time constant [s]
|					
|					To get it right, we'd have to use floating point multiplies. We have no processing power
|					to do this whatsoever, so we approximate the multiplies using shifts.
|					
|					In our case, we have dT = 1/(2.5KHz) & Tau = 1/(100Hz) so a ~ 0.96. However, we need 2^n
|					numbers, so we use a = 240/256 = (256 - 16)/256 = 1 - 16 = 0.9375. This is nice because it also
|					gives us (1-a) = 256/256 - 240/256 = 16/256 = 1/16 = 0.0625. This gives a tau of 161Hz
|					which seems ok.
|					
|					y(i)'s are really 12bit numbers represented by 32 bit numbers in a (16bits.16bits) format.
|					This means we set a pseudo-"decimal point" just to the right of the lower 16bits of the
|					32bits.
|
|					For more information, see the MATLAB file "LPF2.M".
|
| Expects:		Interrupts to be enabled.
|					
|   Date   | Notes
|----------|---------------------------------------------------------------------------------------
| 06/10/01 | Created, based on LV1b & IPRS code.
\--------------------------------------------------------------------------------------------------*/

void LPFSamples (void)
{
	static unsigned long Temp1;
	static unsigned long Temp2;
	static unsigned long Temp3;
	
	PORTE |= (1<<1);	// PULSE RE1 FOR LPF TIME MEASUREMENT
	
	// First, we multiply by 240/256 which is the same as multiplying by (1-1/16). So
	// that's y(i), in the 16.16 form, subtracted by a 4bit right shifted y(i). Note that
	// we ignore the "dropped" nibble since it's 16bits past the "decimal". 

	// Then add (1-a)*x. Normally we'd left shift the 12bits of the sample data
	// by 16bits to get it into the 16.16 form. But, since we're going to then right shift
	// it 4bits in order to multiply by 1/16, we'll just to the final thing which is
	// left shift by only 12bits.

	// This section is the time constraint on the whole damn system, which is
	// why it's unrolled and uses static temporary variables.
			
	if (LPFMask & (1<<0))
	{
		Temp1 = (Y[0] >> 4);
		Temp2 = Y[0] - Temp1; 
		Temp1 = Sample[0];
		Temp3 = Temp1 << 12;
		Y[0] = Temp2 + Temp3;
		LPFMask &= ~(1<<0);	
	}
	if (LPFMask & (1<<1))
	{
		Temp1 = (Y[1] >> 4);
		Temp2 = Y[1] - Temp1; 
		Temp1 = Sample[1];
		Temp3 = Temp1 << 12;
		Y[1] = Temp2 + Temp3;
		LPFMask &= ~(1<<1);	
	}
	if (LPFMask & (1<<2))
	{
		Temp1 = (Y[2] >> 4);
		Temp2 = Y[2] - Temp1; 
		Temp1 = Sample[2];
		Temp3 = Temp1 << 12;
		Y[2] = Temp2 + Temp3;
		LPFMask &= ~(1<<2);	
	}
	if (LPFMask & (1<<3))
	{
		Temp1 = (Y[3] >> 4);
		Temp2 = Y[3] - Temp1; 
		Temp1 = Sample[3];
		Temp3 = Temp1 << 12;
		Y[3] = Temp2 + Temp3;
		LPFMask &= ~(1<<3);	
	}
	if (LPFMask & (1<<4))
	{
		Temp1 = (Y[4] >> 4);
		Temp2 = Y[4] - Temp1; 
		Temp1 = Sample[4];
		Temp3 = Temp1 << 12;
		Y[4] = Temp2 + Temp3;
		LPFMask &= ~(1<<4);	
	}
	if (LPFMask & (1<<5))
	{
		Temp1 = (Y[5] >> 4);
		Temp2 = Y[5] - Temp1; 
		Temp1 = Sample[5];
		Temp3 = Temp1 << 12;
		Y[5] = Temp2 + Temp3;
		LPFMask &= ~(1<<5);	
	}
	if (LPFMask & (1<<6))
	{
		Temp1 = (Y[6] >> 4);
		Temp2 = Y[6] - Temp1; 
		Temp1 = Sample[6];
		Temp3 = Temp1 << 12;
		Y[6] = Temp2 + Temp3;
		LPFMask &= ~(1<<6);	
	}
		
	PORTE &= ~(1<<1);	// PULSE RE1 FOR LPF TIME MEASUREMENT

}




/*-------------------------------------------------------------------------------------------------\
| Function:		SendSamples
|
| Purpose:		Load the samples - raw or LPFed - into the serial queue.
|
| Discussion:	Unrolled to save processing power!
|
| Expects:		Interrupts to be enabled.
|					
|   Date   | Notes
|----------|---------------------------------------------------------------------------------------
| 06/10/01 | Created.
\--------------------------------------------------------------------------------------------------*/

void SendSamples (void)
{
	// Send the header if called for
	
	if (SendMask & (1<<0))
	{
		LoadSerialQueue(0x49); 	// 0x49= ASCII 'i'
	}

	// Send the data if called for
	
	if (SendMask & (1<<1))
	{
		if (MasterSampleMask & (1<<0))
		{
			if (MasterLPFMask & (1<<0))
			{
				LoadSerialQueue( Y[0] >> 16 );		// Low  byte of LPF data
				LoadSerialQueue( Y[0] >> 24 );		// High byte of LPF data
			}
			else
			{
				LoadSerialQueue( Sample[0] );			// Low  byte of raw data
				LoadSerialQueue( Sample[0] >> 8);	// High byte of raw data
			}
		}
		
		if (MasterSampleMask & (1<<1))
		{
			if (MasterLPFMask & (1<<1))
			{
				LoadSerialQueue( Y[1] >> 16 );		// Low  byte of LPF data
				LoadSerialQueue( Y[1] >> 24 );		// High byte of LPF data
			}
			else
			{
				LoadSerialQueue( Sample[1] );			// Low  byte of raw data
				LoadSerialQueue( Sample[1] >> 8);	// High byte of raw data
			}
		}
		
		if (MasterSampleMask & (1<<2))
		{
			if (MasterLPFMask & (1<<2))
			{
				LoadSerialQueue( Y[2] >> 16 );		// Low  byte of LPF data
				LoadSerialQueue( Y[2] >> 24 );		// High byte of LPF data
			}
			else
			{
				LoadSerialQueue( Sample[2] );			// Low  byte of raw data
				LoadSerialQueue( Sample[2] >> 8);	// High byte of raw data
			}
		}
		
		if (MasterSampleMask & (1<<3))
		{
			if (MasterLPFMask & (1<<0))
			{
				LoadSerialQueue( Y[3] >> 16 );		// Low  byte of LPF data
				LoadSerialQueue( Y[3] >> 24 );		// High byte of LPF data
			}
			else
			{
				LoadSerialQueue( Sample[3] );			// Low  byte of raw data
				LoadSerialQueue( Sample[3] >> 8);	// High byte of raw data
			}
		}
		
		if (MasterSampleMask & (1<<4))
		{
			if (MasterLPFMask & (1<<4))
			{
				LoadSerialQueue( Y[4] >> 16 );		// Low  byte of LPF data
				LoadSerialQueue( Y[4] >> 24 );		// High byte of LPF data
			}
			else
			{
				LoadSerialQueue( Sample[4] );			// Low  byte of raw data
				LoadSerialQueue( Sample[4] >> 8);	// High byte of raw data
			}
		}

		if (MasterSampleMask & (1<<5))
		{
			if (MasterLPFMask & (1<<5))
			{
				LoadSerialQueue( Y[5] >> 16 );		// Low  byte of LPF data
				LoadSerialQueue( Y[5] >> 24 );		// High byte of LPF data
			}
			else
			{
				LoadSerialQueue( Sample[5] );			// Low  byte of raw data
				LoadSerialQueue( Sample[5] >> 8);	// High byte of raw data
			}
		}
		
		if (MasterSampleMask & (1<<6))
		{
			if (MasterLPFMask & (1<<6))
			{
				LoadSerialQueue( Y[6] >> 16 );		// Low  byte of LPF data
				LoadSerialQueue( Y[6] >> 24 );		// High byte of LPF data
			}
			else
			{
				LoadSerialQueue( Sample[6] );			// Low  byte of raw data
				LoadSerialQueue( Sample[6] >> 8);	// High byte of raw data
			}
		}
	}

	// Send the CRC if required
	
	if (SendMask & (1<<1))
	{
		LoadSerialQueue( CRC );			// Send CRC low  byte (calculated in LoadSerialQueue)
		LoadSerialQueue( CRC >> 8 );	// Send CRC high byte (calculated in LoadSerialQueue)
	}

	// Clear the flag
	
	FlgSendSamples = 0;
}





/*-------------------------------------------------------------------------------------------------\
| Function:  	ExecRcvdCmd
|
| Purpose:   	Process a command that was received over the UART from the PC.
|
| Discussion:	
|
| Expects:		-
|					
|   Date   | Notes
|----------|---------------------------------------------------------------------------------------
| 06/11/00 | Created.
\--------------------------------------------------------------------------------------------------*/

void	ExecRcvdCmd (void)
{
	char *MemPtr;
	
	// First, see if we should process the command
	
	if (RxQueue[0] == 'T')
	{
		if (RxQueue[1] == 'T' && RxQueue[2] == 0 )
		{
			LoadSerialQueue( 'T' );						// 'T'
			LoadSerialQueue( RxQueue[3] );				// Sequence Low
			LoadSerialQueue( RxQueue[4] );				// Sequence High
			LoadSerialQueue( 5 );							// Length of data = 5 bytes
			LoadSerialQueue( TimeStampL );				// TimeStamp byte 1 (LSB)
			LoadSerialQueue( TimeStampH );				// TimeStamp byte 2
			LoadSerialQueue( TimeStampH >>  8 );		// TimeStamp byte 3
			LoadSerialQueue( TimeStampH >> 16 );		// TimeStamp byte 4
			LoadSerialQueue( TimeStampH >> 24 );		// TimeStamp byte 5 (MSB)
		}
	}
	
	else if (RxQueue[0] == 'I')
	{
		if (RxQueue[1] == 'R' && RxQueue[2] == 2 )
		{
			// READ COMMAND- SEND REGISTER OUT QUEUE
		}
		else if (RxQueue[1] == 'W' && RxQueue[2] == 3 )
		{

			// Write to a register in memory. Use a pointer to do that - but the pointer
			// doesn't specify the BANK, which should be specified by the IRP Bit.
			// DOES IRP REALLY AFFECT THE POINTER?!?!
			
			if	(RxQueue[6])
				IRP = 1;
			else
				IRP = 0;

			MemPtr = RxQueue[5];
			*MemPtr = RxQueue[7];
		}

	}
	// Reset the received command flag so we know we processed this in time.

	FlgRecvdCmd = 0;
}





/*-------------------------------------------------------------------------------------------------\
| Function:		LoadSerialQueue
|
| Purpose:		Loads a byte into the UART (serial port) queue.
|
| Discussion:	The SerialQueue[] (69 bytes long in Bank 2) is a ring FIFO queue that contains bytes
|					that are to be sent out of UART.
|             
| Expects:		To be called by any routine wanting to send a serial message.
|					The message to be sent is stored in the char passed in, Byte
|
|   Date   | Notes
|----------|---------------------------------------------------------------------------------------
| 06/10/01 | Created using old code.
\-------------------------------------------------------------------------------------------------*/

void LoadSerialQueue (char Byte)
{


	// If there was a message queue overflow in the last attempt to store a message, then we swap
	// the current message trying to be saved with the "error queue overflow" message. We keep trying
	// this until this error message finally manages to get into the queue. We then wait for 256 messages
	// stored in the queue without an overflow before we send the error message again (see below).

	if (SystemError)
		Byte = SystemError;	// Send out the System Error if there is one.
			
	// Save the first byte of the message in the Message Queue and check for overflow

	SerialQueue[SerialQIn] = Byte;

	++SerialQIn;
	if (SerialQIn >= SERIALQLENGTH)
		SerialQIn = 0;

	if (SerialQIn == SerialQOut)
	{
		// The index in equals the index out so we have a ring buffer overflow condition.
		// Just throw an error for now.

		SystemError = 5;
	}

	// No matter what, we've just loaded a byte into the queue so we need to turn on the transmit interrupts

	TXIE = 1;
}



