/*-------------------------------------------------------------------------------------------------\
| IMU 1b Microcontroller Firmware
|
| File:					IMUCINI.C		Initialization Code
|
| Global Routines:	Initialize()
|
| Local Routines:	   None
|
\-------------------------------------------------------------------------------------------------*/

#include	<pic.h>
#include "imuc.h"

//-------------------------------------------------------------------------------------------------
//  Define Locally Global Variables (variables accessible by functions in this file only)
//-------------------------------------------------------------------------------------------------

	// None

//-------------------------------------------------------------------------------------------------	
//  Local #defines
//-------------------------------------------------------------------------------------------------

	// None

//-------------------------------------------------------------------------------------------------	
//  Local Function Prototypes
//-------------------------------------------------------------------------------------------------

	// None

/*-------------------------------------------------------------------------------------------------\
| Function:   Initialize
|
| Purpose:    Initialize the Ports, Peripherals and variables
|
| Discussion: All sorts of ports and peripherals need to be initialized and turned on after
|				  a reset (whether it was a MCLR or power up reset).
|             
| Expects:    To be called after any type of reset.
|             Interrupts to be disabled.
|					
|   Date   | Notes
|----------|---------------------------------------------------------------------------------------
| 06/06/01 | Created based on previous code. 
\--------------------------------------------------------------------------------------------------*/

void Initialize (void)
{

	char	i;
	
	// First we need to bring all of the PORTs into their correct state. 

	// Note that we set the PORT registers FIRST, and then the TRIS registers. In this order,
	// we can be sure that the output pins come "up" out of tri-state in the right state.

	PORTA = 0;		// Note: ADCON1 also affects PORTA and is set below in the ADC section

	TRISA = 0b00100001;
			 // ||||||||
			 // |||||||+----------- In  <- RA0/AN0 = IMUTEMP thermister input
		    // ||||||+------------ Out -> RA1/AN1 = Unused
		    // |||||+------------- Out -> RA2/AN2 = Unused
		    // ||||+-------------- Out -> RA3/AN3 = Unused
		    // |||+--------------- Out -> RA4/T0I = Unused (Output open collector)
			 // ||+---------------- Out -> RA5/AN4 = Unused
		    // |+----------------- (Not implemented in PIC)
	       // +------------------ (Not implemented in PIC)

	PORTA = 0;		// Rewrite in case of bit flips from TRISA
	
// -------------------------------
	
	RBPU = 0;		// Turn on Port B weak (~60kohm) pull-up resistors	
	TRISB = 0b11111111;
			 // ||||||||
			 // |||||||+----------- In <- RB0/INT = Unused
		    // ||||||+------------ In <- RB1     = Unused
		    // |||||+------------- In <- RB2     = Unused
		    // ||||+-------------- In <- RB3/PGM = Unused
		    // |||+--------------- In <- RB4     = INTADC (Interrupt from MAX197 ADC) - IOC
			 // ||+---------------- In <- RB5     = INTCAD (Interrupt from SJA1000 CAN IC)- IOC
		    // |+----------------- In <- RB6/PGC = MPLAB ICD - IOC
	       // +------------------ In <- RB7/PGD = MPLAB ICD - IOC

// -------------------------------
	
	PORTC = 0b00000111;
			 // ||||||||
			 // |||||||+----------- Out -> RC0 = /RD = off
		    // ||||||+------------ Out -> RC1 = /WR = off
		    // |||||+------------- Out -> RC2 = /HBEN = off
		    // ||||+-------------- Out -> RC3 = ALE = off
		    // |||+--------------- Out -> RC4 = TEST = off
			 // ||+---------------- Out -> RC5 = Debug1 = off
		    // |+----------------- Out -> RC6 = UART Transmit
	       // +------------------ In  <- RC7 = UART Receive
	
	TRISC = 0b10000000;
			 // ||||||||
			 // |||||||+----------- Out -> RC0 = /RD
		    // ||||||+------------ Out -> RC1 = /WR
		    // |||||+------------- Out -> RC2 = /HBEN
		    // ||||+-------------- Out -> RC3 = ALE
		    // |||+--------------- Out -> RC4 = TEST
			 // ||+---------------- Out -> RC5 = Debug
		    // |+----------------- Out -> RC6 = UART Transmit
	       // +------------------ In  <- RC7 = UART Receive

	PORTC = 0b00000111;	// Write again in case of bit flips from the TRISC command.


	// -------------------------------

	// Set up PORTE First so that the peripheral chips are guaranteed to be offline
	// when the data bus comes up.
	
	PORTE = 0b00000011;
			 // ||||||||
			 // |||||||+----------- RE0 = /CS0 = off
		    // ||||||+------------ RE1 = /CS1 = off
		    // |||||+------------- RE2 = Debug2 LED = off
		    // ||||+-------------- (Not implemented in PIC)
		    // |||+--------------- (Not Implemented in PIC)
			 // ||+---------------- (Not Implemented in PIC)
		    // |+----------------- (Not Implemented in PIC)
	       // +------------------ (Not Implemented in PIC)

	TRISE = 0b00000000;
			 // ||||||||
			 // |||||||+----------- Out -> RE0 = /CS0 to MAX197 ADC
		    // ||||||+------------ Out -> RE1 = /CS1 to SJA1000
		    // |||||+------------- Out -> RE2 = Debug2 LED
		    // ||||+-------------- (Not Implemented in PIC)
		    // |||+--------------- PSPMode = 0 (GPIO)
			 // ||+---------------- NA     IBOV (Read Only)
		    // |+----------------- NA     OBF  (Read Only)
	       // +------------------ NA     IBF  (Read Only)

	PORTE = 0b00000011;	// Write again in case of bit flips from the TRISE command.
	
	// -------------------------------
	
	PORTD = 0;		// PortD is the Data Bus - off for now
	
	TRISD = 0;		// Port D is all outputs (address/data bus)

	PORTD = 0;		// Rewrite in case of bit flips

	// ---------------------------------------------------------------------------

	// Before turning on any peripherals, initialize all relevant data so that
	// the timer won't time out, or the receive buffer won't overflow, etc., as we
	// initialize the data.
	
	FlgPowerUp = 1;

	MasterSampleMask = 0b01111111;	// Sample all sensors
	MasterLPFMask = 0b01111111;		// LPF all sensors
	SendSamplesCnt = 8;					// 1/(2.5KHz) * 8 = 3.2ms = 312.5Hz
	SendMask	= 0b00000111;				// .0 = Send header, .1 = send data, .2 = send CRC

	// ---------------------------------------------------------------------------

	// Bring the peripherals online. This means Timer2, the UART and the 10bit ADC.

	// Initialize Timer2 to make a ??? interrupt given Fosc/4 = 5MHz.

	//  2.5KHz: T2CON =           , PR2 =  99 -> 1:4,  99, 1:5
	//    1KHz: T2CON = 0b00101101, PR2 = 250 -> 1:4, 250, 1:5
	// 312.5Hz: T2CON = 0b01010110, PR2 =  94 -> 1:16, 100, 1:10
	
	T2CON = 0b01010110;
			 // ||||||||
			 // |||||||+----------- T2CKPS0 = 01   (1:16 Prescaler)
		    // ||||||+------------ T2CKPS1
		    // |||||+------------- TMR2ON  = 1    (Timer2 is on)
		    // ||||+-------------- TOUTPS0 = 0101 (1:10 Postscaler)
		    // |||+--------------- TOUTPS1
			 // ||+---------------- TOUTPS2
		    // |+----------------- TOUTPS3
	       // +------------------ (Not Implemented in PIC)
	
	PR2 = 31;		// Period in ticks of Fosc/4
	
	// Initialize the ADC and PORTA for the ADC channels.
	
	ADCON1 =	0b10001110;
			//   ||||||||
			//	  |||||||+----------- PCFG0 = 1110 (Only RA0 is ADC in (AN0))
			//	  ||||||+------------ PCFG1
			//	  |||||+------------- PCFG2
			//	  ||||+-------------- PCFG3
			//	  |||+--------------- (Not Implemented in PIC)
			//	  ||+---------------- (Not Implemented in PIC)
			//	  |+----------------- (Not Implemented in PIC)
			//	  +------------------ ADFM = 1     (Right justified: ADRESH = ADC9:8, ADRESL = ADC7:0)

	ADCON0 = 0b10000001;
			//   ||||||||	
			//   |||||||+----------- ADON  = 1   (Turn the ADC on)
			//   ||||||+------------ (Not Implemented in PIC)
		  	//   |||||+------------- GO    = 0   (Don't start a conversion)
			//   ||||+-------------- CHS0  = 000 (Start with Channel 0)
			//   |||+--------------- CHS1
			//   ||+---------------- CHS2
			//   |+----------------- ADCS0 = 10  (Fosc/32)
			//   +------------------ ADCS1

	// Initialize the UART for receive and transmission at 57.6kbps given Fosc = 20MHz

	SPBRG = 21;	// Set the bps rate to 57.6kbps given BRGH = 1 -> Acutally is 56.818kbps (1.36% slow)
	
	RCSTA = 0b10010000;
			//  ||||||||	
			//  |||||||+----------- RX9D
			//  ||||||+------------ OERR
			//  |||||+------------- FERR
			//  ||||+-------------- ADDEN = 0
			//  |||+--------------- CREN  = 1
			//  ||+---------------- SREN
			//  |+----------------- RX9   = 0
			//  +------------------ SPEN  = 1

	TXSTA = 0b00100100;
			//  ||||||||	
			//  |||||||+----------- TX9D  (unused: 9th bit) 
			//  ||||||+------------ TRMT
			//  |||||+------------- BRGH = 1 (high speed)
			//  ||||+-------------- Unimplemented
			//  |||+--------------- SYNC = 0 (asynchronous)
			//  ||+---------------- TXEN = 1 (enable transmission)
			//  |+----------------- TX9D = 0 (8 data bits)
			//  +------------------ CSRC
	
	// ---------------------------------------------------------------------------

	// Clear any pending interrupts

	i = PORTB;
	RBIF = 0;
	
	TMR2IF = 0;				// Clear the Timer2 interrupt
	
	i = RCREG;				// Move the RCREG into the W register in order to clear the interrupt flag
	CREN = 1;				// Clear any overrun errors in the UART receive hardware
	
	// Turn on all of the interrupt enables and finally the global enables.

	RBIE = 1;				// Enable PORTB interrupt on change, Bits 4-7
	RCIE = 1;				// Enable UART interrupt on receive buffer full
	TMR2IE = 1;				// Enable interrupt on Timer2 to PR2 match

	PEIE = 1;				// Enable peripheral interrupts
	GIE = 1;					// Enable global interrupts

}



