/*-------------------------------------------------------------------------------------------------\
| IMU 1b Microcontroller Firmware
|
| File:					IMUCINT.C		Interrupt Routines
|
| Global Routines:	ISR()
|
| Local Routines:		Timer2()
|							Transmit()
|							Receive()
|							PortBIOC()
|							SampleADC()
|
| IMPORTANT NOTE ON THE STRUCTURE OF THIS MODULE:
| -----------------------------------------------
| PICC will automatically save the context of all functions that an "interrupt"-labeled routine
| (ISR below) calls. However, the functions called must be ABOVE THE INTERRUPT-LABELED FUNCTION
| in the same code module (i.e., the same file). Thus, this module must RETAIN THE ORDER OF THE
| FUNCTIONS with the interrupt function being last in the module. For more information, see the
| Hi-Tech PIC C Manual.
\-------------------------------------------------------------------------------------------------*/

#include	<pic.h>
#include "imuc.h"


//-------------------------------------------------------------------------------------------------
//  Define Locally Global Variables (variables accessible by functions in this file only)
//-------------------------------------------------------------------------------------------------

	// None

//-------------------------------------------------------------------------------------------------
//  Local #defines
//-------------------------------------------------------------------------------------------------

	// None

//-------------------------------------------------------------------------------------------------	
//  Local Function Prototypes
//-------------------------------------------------------------------------------------------------

	void	SampleADC	(void);
	void	Timer2		(void);
	void	PortBIOC		(void);
	void	Receive		(void);
	void	Transmit		(void);



/*-------------------------------------------------------------------------------------------------\
| Function:  	SampleADC
|
| Purpose:   	Sample and low pass filter the IMU sensors via the MAX197ADC
|
| Discussion:	SampleADC is an interrupt-based function which is called by two different
|					interrupts: Timer0 and PORTBIOC (Port B Interrupt On Change).
|					SensorNum = 0 -> SensorMask.0 = 1 -> Sample[0] = X
|					SensorNum = 1 -> SensorMask.1 = 1 -> Sample[1] = y
|					SensorNum = 2 -> SensorMask.2 = 1 -> Sample[2] = z
|					SensorNum = 3 -> SensorMask.3 = 1 -> Sample[3] = xy
|					SensorNum = 4 -> SensorMask.4 = 1 -> Sample[4] = phi
|					SensorNum = 5 -> SensorMask.5 = 1 -> Sample[5] = psy
|					SensorNum = 6 -> SensorMask.6 = 1 -> Sample[6] = theta
|
| Timing:		State 0 is about 30Tcyc, State 1 is about 40Tcyc.
|
| Expects:		Interrupts to be disabled because we're in an interrupt.
|					
|   Date   | Notes
|----------|---------------------------------------------------------------------------------------
| 06/06/01 | Created based on previous code. 
\--------------------------------------------------------------------------------------------------*/

void SampleADC (void)
{
	static char SensorNum;		// 0 - 6 -> which sensor we're on
	static char SampleState;	// state = 0 -> Running the ADC, 1 -> Reading ADC values

	char	Temp;
	
	// Globals used: SampleMask, LPFMask, MasterLPFMask


	// This is a little backwards; we read the data first and then turn on the ADC. The reason is
	// because we can save an interrupt and all of the overhead if we immediately turn the
	// ADC on after we grab the last data. Since SampleState is intialized to 0, we start off
	// correctly so the order doesn't matter!
	
	if (SampleState == 1)
	{
		// Send the command to the ADC to begin capturing the data

		PORTE &= ~(1<<0);								// Select the ADC (/CE0)
		TRISD = 0xFF;									// Set Port D to read
		PORTC &= ~(1<<2);								// Select the low byte (HBEN)
		PORTC &= ~(1<<0);								// Select Read to the ADC (/RD)
		Temp = PORTD;									// Read in the low byte data
		PORTC |=  (1<<2);								// Select the high byte (HBEN)
		Sample[SensorNum] = ((PORTD & 0x0F)<<8) + Temp;	// Read in the high byte of the data
		PORTC |=  (1<<0);								// Deselect Read to ths ADC(/RD)
		TRISD = 0x00;									// Set Port D to write
		PORTE |=  (1<<0);								// Deselect the ADC (/CE0)

		// When we read the ADC we reset the /INT command, which triggers
		// an interrupt on PORTB. We need to clear that interrupt here.
			
		Temp = PORTB;
		RBIF = 0;
		
		// Flag the mainline routines that we have a valid sample ready to be
		// Low pass filtered. Then make sure we should be low pass filtering
		// that sensor.

		LPFMask |= (1<<SensorNum);
		LPFMask &= MasterLPFMask;

		// We're done reading the ADC. Restart the sampling sequence.
		
		SampleState = 0;

		// If all of the Mask bits have been cleared, then flag that we're done sampling
		
		if (!SampleMask)
			FlgSampleDone = 1;
	}


	if (SampleState == 0)
	{
		if (SampleMask & 1<<0)
		{
			SampleMask &= ~(1<<0);		// clear the current sensor mask bit
			SensorNum = 0;					// Set the current sensor value
			Temp = 0b01000010; 			// MAX197A Command: Normal, Int clock and acq., 0-5V, CH2 = X
		}
		else if (SampleMask & 1<<1)
		{
			SampleMask &= ~(1<<1);		// clear the current sensor mask bit
			SensorNum = 1;					// Set the current sensor value
			Temp = 0b01000111; 			// MAX197A Command: Normal, Int clock and acq., 0-5V, CH7 = Y
		}
		else if (SampleMask & 1<<2)
		{
			SampleMask &= ~(1<<2);		// clear the current sensor mask bit
			SensorNum = 2;					// Set the current sensor value
			Temp = 0b01000000; 			// MAX197A Command: Normal, Int clock and acq., 0-5V, CH0 = Z
		}
		else if (SampleMask & 1<<3)
		{
			SampleMask &= ~(1<<3);		// clear the current sensor mask bit
			SensorNum = 3;					// Set the current sensor value
			Temp = 0b01000011; 			// MAX197A Command: Normal, Int clock and acq., 0-5V, CH3 = XY
		}
		else if (SampleMask & 1<<4)
		{
			SampleMask &= ~(1<<4);		// clear the current sensor mask bit
			SensorNum = 4;					// Set the current sensor value
			Temp = 0b01000100; 			// MAX197A Command: Normal, Int clock and acq., 0-5V, CH4 = Phi - Around X
		}
		else if (SampleMask & 1<<5)
		{
			SampleMask &= ~(1<<5);		// clear the current sensor mask bit
			SensorNum = 5;					// Set the current sensor value
			Temp = 0b01000101; 			// MAX197A Command: Normal, Int clock and acq., 0-5V, CH5 = Psy - Around Y
		}
		else if (SampleMask & 1<<6)
		{
			SampleMask &= ~(1<<6);		// clear the current sensor mask bit
			SensorNum = 6;					// Set the current sensor value
			Temp = 0b01000001;			// MAX197A Command: Normal, Int clock and acq., 0-5V, CH1 = Theta - Around Z
		}

		// Send the command to the ADC to begin capturing the data

		PORTD = Temp;				// Throw the command on the Data bus (Port D)
		PORTE &= ~(1<<0);			// Select the ADC (/CE0)
		PORTC &= ~(1<<1);			// Select Write to the ADC (/WR)
		PORTC |=  (1<<1);			// Deselect Write to ths ADC(/WR)
		PORTE |=  (1<<0);			// Deselect the ADC (/CE0)
			
		SampleState = 1;
	}
}



/*-------------------------------------------------------------------------------------------------\
| Function:  	Timer2
|
| Purpose:   	Interrupt at some known rate. 
|
| Discussion:	-
|
| Expects:		Interrupts to be disabled.
|					
|   Date   | Notes
|----------|---------------------------------------------------------------------------------------
| 06/06/01 | Created based on previous code. 
\--------------------------------------------------------------------------------------------------*/

void Timer2 (void)
{

	static char	SendSamplesClk;

	PORTC |= (1<<3); // PULSE RC3 FOR CHECKING TIMER2 TIME

	// Clear the Timer2 interrupt & increment the global time
	
	TMR2IF = 0;

	++TimeCnt;

	// Increment the UART "send sample" count and flag the mainline routine when ready to send a
	// packet of data.

	++SendSamplesClk;

	if (SendSamplesClk >= SendSamplesCnt)
	{
		SendSamplesClk = 0;
		FlgSendSamples = 1;
	}
	
	// Check for timing / overrun errors

	if (SampleMask)
		SystemError = 1;
	else if (LPFMask)
		SystemError = 6;
	
	// Set up the variables for the SampleADC routine, and let the world know we're sampling.

	FlgSampleDone = 0;
	SampleMask = MasterSampleMask;

	// Go sample, & then return & return from interrupt
	
	SampleADC();

	PORTC &= ~(1<<3);	// PULSE FOR CHECKING TIMER2 TIME


}


/*-------------------------------------------------------------------------------------------------\
| Function:  	PortBIOC
|
| Purpose:   	To pick up the ADC and CAN interrupts
|
| Discussion:	-
|
| Expects:		Interrupts to be disabled.
|					
|   Date   | Notes
|----------|---------------------------------------------------------------------------------------
| 06/09/01 | Created. 
\--------------------------------------------------------------------------------------------------*/

void PortBIOC (void)
{

	char i;
	
	// Clear the PortBIOC interrupt. When the CAN controller is implemented, this routine should try
	// and tell the different between the interrupts. For now, all we have implemented is the
	// ADC so that's all we'll consider for this routine.

	i = PORTB;
	RBIF = 0;

	// Go finish sampling the sensors
	
	SampleADC();

	// Return and then return from interrupt
}



/*-------------------------------------------------------------------------------------------------\
| Function:  	Receive
|
| Purpose:   	Grabs the byte in the UART receive buffer and puts it into the received message queue.
|
| Discussion:	Bytes coming in the UART are commands from the PC. This routine takes the bytes and
|					stores them in a queue for processing. Takes a timestamp if a "T" command arrives.
|
|					There are 3 states to the receive buffer. State 0 is when we're looking for a command
|					and ignoring any zero's that are comming in. State 1 is when we're receiving the
|					command which consists of 5 bytes.	State 3 is when we've received the command we're
|					getting any data that was sent.
|
| Expects:		To be called on the UART Receiver buffer full interrupt.
|					Interrupts to be disabled since we're in an interrupt.
|
|   Date   | Notes
|----------|---------------------------------------------------------------------------------------
| 06/06/01 | Created based on previous code. 
\--------------------------------------------------------------------------------------------------*/

void Receive (void)
{
	static char ReceiveState;	
	char			Temp;

	
	// First we clear the interrupt by reading the UART Receive register

		Temp = RCREG;

	if (!ReceiveState)
	{
		// We're looking for the beginning of a command. Ignore all 0's - that's how we'll re-sync
		// our commands
		
		if (Temp == 0x00)
			return;

		// The byte is nonzero. First, check for "overrun" error in the mainline code and stop the
		// mainline code from processing whatever command it was processing.

		if (FlgRecvdCmd)
		{
			SystemError = 3;
			FlgRecvdCmd = 0;
		}
			
		// Start packing the received bytes into the 20 byte receive buffer.

		RxQPtr = 0;
		RxQueue[0] = Temp;

		// Check to see if the first byte of the command is a "T" for timestamp. If it is, then
		// store the 32 byte clock in a register to be used when we build the T command response packet
		// in the mainline command processing routine.

		if (Temp == 'T')
		{
			TimeStampL = TMR2;
			TimeStampH = TimeCnt;

			// This works, but with one exception: if the T2IF flag is set, then the counter has
			// rolled over to zero since we started this routine. If it did, then the T2IF flag
			// will be set - if it is, we set the low byte of the timestamp to FF to be the
			// "best estimate" of the time when we received the "T".
			
			if (TMR2IF)
				TimeStampL = 0xFF;
		}
		
		// Switch to the next state
		
		ReceiveState = 1;
	}

	
	else if (ReceiveState == 2)
	{
		// We're officially reading in a command. Each command has at least 5 bytes:
		// Address: T = Time (both), S = Spintable, I = IMU
		// Command: T = Time, R = Read register, W = Write Register, S = Spintable setting
		// Count:   # of data bytes following command
		// Sequence Number: a 16 bit command sequence number

		++RxQPtr;

		RxQueue[RxQPtr] = Temp;

		// Check to see if we're at the last byte of the command header. If we are, then
		// check the count of the data following the command. If we're at the end and the
		// count = 0, then we're done with this command. Else get the rest of the data.

		if (RxQPtr == 4)
		{
			if(RxQueue[4] == 0)
			{
				FlgRecvdCmd = 1;
				ReceiveState = 0;
			}
			else
				ReceiveState = 3;
		}
	}

	
	else	// (ReceiveState == 3)
	{

		// Now we're reading the data part of the command. Don't read more than 15 bytes of data.
		// Note that we're not going to parse the command here - we'll do that in the mainline code.
		
		++RxQPtr;

		RxQueue[RxQPtr] = Temp;

		if	(RxQPtr >= (RxQueue[4] + 4) || RxQPtr >= 19)
		{
			FlgRecvdCmd = 1;
			ReceiveState = 0;
		}
	}
}



/*-------------------------------------------------------------------------------------------------\
| Function:  	Transmit
|
| Purpose:   	Stuff the UART Transmit buffer with data from the Serial Queue until there is no
|					data left in the queue.
|
| Discussion:	The "Serial Queue" in Bank 2 contains 69 bytes and is used as a FIFO ring buffer in an
|					array (with index 0 to 68). 69 bytes stores 23 messages, each of which is queued up in
|					this buffer. This	routine grabs a byte from the queue and sends it out the UART at 9600bps
|					(960Bps). If there are no more bytes in the queue, it'll turn off the UART TX buffer empty
|					interrupt enable. The transmit interrupts are re-enabled by any call to the LoadSerialQueue
|					routine which loads the queue and turns on the interrupts.
|
|					Note that this routine reads the serial queue and blindly sends out the data - it does NO
|					formatting of the outgoing messages.
|
| Expects:		To be called on the UART TX buffer empty interrupt.
|					Interrupts to be disabled.
|					
|   Date   | Notes
|----------|---------------------------------------------------------------------------------------
| 06/06/01 | Created based on previous code. 
\--------------------------------------------------------------------------------------------------*/
   
void Transmit (void)
{
	// If this routine was called, then LoadSerialQueue() has enabled the "Interrupt on TX Buffer Empty"
	// interrupts (TXIF) meaning that there should be unread data in the Serial Queue.

	// Check if there are in fact unread messages
	
	if (SerialQOut == SerialQIn)
	{
		// There are no message left. Turn off interrupts and exit.
		TXIE = 0;
	}

	else
	{
		// There are messages; send the next byte out the UART

		TXREG = SerialQueue[SerialQOut];

		// Increment the message pointer and reset to the beginning of the queue if necessary
	
		++SerialQOut;
	
		if (SerialQOut >= SERIALQLENGTH)
			SerialQOut = 0;
	}
}





/*-------------------------------------------------------------------------------------------------\
| Function:  	ISR
|
| Purpose:   	ISR Is the interrupt service routine. It calls the ISR's for
|					each type of interrupt.
|
| Discussion:	There are three kinds of interrupts in the IMUC: Timer2 at 1KHz, RCIF which occurs
|					when we receive a byte from the UART, and TXIF when we are ready to send a byte
|					out of the UART (which is only enabled when there is something in the serial
|					queue to send out). This routine calls each one according to the interrupt flag
|					set. It also determines the interrupt priority by the order in which they're called.
|
| Expects:		To be called on any kind of PIC16F877 interrupt.
|					For global interupts to be disabled by the interrupt (INTSTA, GIE = 0).
|					Hi-Tech C to save and restore all variables used in and by these int. routines.
|					
|   Date   | Notes
|----------|---------------------------------------------------------------------------------------
| 06/06/01 | Created based on previous code. 
\--------------------------------------------------------------------------------------------------*/

void interrupt ISR (void)
{
	// Turn off the Mainline LED since we're in the ISR

		PORTC &= ~(1<<5);

	// Now we determine who interrupted us and go to that ISR. Note that the order of checking
	// the interrupt flags determines the interrupt priority.
	
		if (RBIF)				// Check PortBIOC for an interrupt on change of state
			PortBIOC();

		else if (RCIF)			// Check the UART Receive Buffer Full interrupt
			Receive();

		else if (TMR2IF)		// Check the Timer2 interrupt
			Timer2();

		else if (TXIF)			// Check the UART Transmit Buffer Empty interrupt
			Transmit();

		else
			SystemError = 4;	// Unknown interrupt


	// Turn on the Mainline LED since we're exiting the ISR

		PORTC |= (1<<5);

	// Note: This routine Automatically RETFIEs (which enables interrupts)
}

