/*-------------------------------------------------------------------------------------------------\
| IMU 1b Microcontroller Firmware
|
| File:					IMUC_MSC.C		Misc. Functions
|
| Global Routines:	LoadSerialQueue(char,char,char)
|							ProcessRcvdMsgs()
| 
| Local Routines:		None
|							
\-------------------------------------------------------------------------------------------------*/

#include	<pic.h>
#include "imuc.h"

//-------------------------------------------------------------------------------------------------
//  Define Locally Global Variables (variables accessible by functions in this file only)
//-------------------------------------------------------------------------------------------------

	// None

//-------------------------------------------------------------------------------------------------	
//  Local #defines
//-------------------------------------------------------------------------------------------------

	// None

//-------------------------------------------------------------------------------------------------	
//  Local Function Prototypes
//-------------------------------------------------------------------------------------------------

	// None


/*-------------------------------------------------------------------------------------------------\
| Function:		LoadSerialQueue
|
| Purpose:		Loads a message stored in Msg0:Msg1:Msg2 into the SerialQueue[]
|
| Discussion:	The SerialQueue[] (69 bytes long in Bank 2) is a ring FIFO queue that contains messages
|					that are to be sent out of the Logic Module's UART to the rest of the RAIII system
|					(usually just the SetBox; in the future, it could be Flaps Displays or whatnot). This
|					routine loads in 3 bytes from Msg0:2 and tries to store them into the queue while
|					looking for pointer (Index, really)	overflows and queue overflows.
|
|					On queue overflows, two flags are set and the next successfully queued message is the
|					"queue overflow error" message which guaratees that the external world finds out about
|					the error. The error message is only sent once - it takes 256 non-overflowed queued
|					messages to send the error message again.
|             
| Expects:		To be called by any routine wanting to send a serial message.
|					The message to be sent is stored in Msg0, Msg1 and Msg2.
|
|   Date   | Notes
|----------|---------------------------------------------------------------------------------------
| 03/23/01 | Synced LM and SB functions.
| 03/10/01 | Fixed fence post bug involving SerialQLength.
| 11/11/00 | Moved to HiTech C
| 10/26/00 | Moved to 3 bytes in at a time in Msg0:2 to simplify. Completed Message
|          | Queue overflow handling. Documented.
| 10/13/00 | Created.
\-------------------------------------------------------------------------------------------------*/

void LoadSerialQueue (char Msg0, char Msg1, char Msg2)
{

	static	bit	FlgSerialQueueOk;
	static	bit	FlgSendOFErrMsg;
	static	char	SerialQNoOFCount;
	
	// First, turn off the UART interrupts because we don't want TxInt reading bytes from the buffer
	// while we're loading bytes into the buffer: weird things could happen, such as the in and out
	// pointer being equal right after we increment the in ptr. We don't mind other interrupts, like
	// the RcInt or Sample interrupts, however. To be safe, we do this twice in order to
	// guarantee that if we get this right on a TX interrupt we clear the flag when we come back.

	TXIE = 0;
	TXIE = 0;

	// If there was a message queue overflow in the last attempt to store a message, then we swap
	// the current message trying to be saved with the "error queue overflow" message. We keep trying
	// this until this error message finally manages to get into the queue. We then wait for 256 messages
	// stored in the queue without an overflow before we send the error message again (see below).

	if (FlgSendOFErrMsg)
	{
		Msg0 = 0x00;  // Overwrite whatever the current message is with the error message
		Msg1 = 0x20;  // (flag cleared below after this message has been stored in the queue)
		Msg2 = 0x01;
	}
	
	// Save the first byte of the message in the Message Queue and check for overflow

	SerialQueue[SerialQIn] = Msg0;

	++SerialQIn;
	if (SerialQIn >= SERIALQLENGTH)
		SerialQIn = 0;

	if (SerialQIn == SerialQOut)
	{
		// The index in equals the index out so we have a ring buffer overflow condition.
		// Back up the pointer and signal the error condition
		if (SerialQIn == 0)
			SerialQIn = (SERIALQLENGTH - 1);
		else
			--SerialQIn;
			

		// If we don't already know about the error, we'll clear the queue ok flag and set the flag
		// to send the error message.
			
		if (FlgSerialQueueOk)
		{
			FlgSerialQueueOk = 0;
			FlgSendOFErrMsg = 1;
		}
		// We've had a buffer overflow, so obviously there are messages in the queue, so we'll enable
		// the UART transmit interrupts as we exit
		TXIE = 1;
		return;		// EXIT
	}

	// Save the second byte of the message in the Message Queue and check for overflow

	SerialQueue[SerialQIn] = Msg1;

	++SerialQIn;
	if (SerialQIn >= SERIALQLENGTH)
		SerialQIn = 0;

	if (SerialQIn == SerialQOut)
	{
		// The index in equals the index out so we have a ring buffer overflow condition.
		// Back up the pointer past previous bytes and signal the error condition
		if (SerialQIn == 0)
			SerialQIn = (SERIALQLENGTH - 1);
		else
			--SerialQIn;
		if (SerialQIn == 0)
			SerialQIn = (SERIALQLENGTH - 1);
		else
			--SerialQIn;

		// If we don't already know about the error, we'll clear the queue ok flag and set the flag
		// to send the error message.
			
		if (FlgSerialQueueOk)
		{
			FlgSerialQueueOk = 0;
			FlgSendOFErrMsg = 1;
		}
		// We've had a buffer overflow, so obviously there are messages in the queue, so we'll enable
		// the UART transmit interrupts as we exit
		TXIE = 1;
		return;		// EXIT
	}

	// Save the Third byte of the message in the Message Queue and check for overflow

	SerialQueue[SerialQIn] = Msg2;

	++SerialQIn;
	if (SerialQIn >= SERIALQLENGTH)
		SerialQIn = 0;

	if (SerialQIn == SerialQOut)
	{
		// The index in equals the index out so we have a ring buffer overflow condition.
		// Back up the pointer past previous 3 bytes and signal the error condition
		
		if (SerialQIn == 0)
				SerialQIn = (SERIALQLENGTH - 1);
			else
				--SerialQIn;
		if (SerialQIn == 0)
				SerialQIn = (SERIALQLENGTH - 1);
			else
				--SerialQIn;
		if (SerialQIn == 0)
				SerialQIn = (SERIALQLENGTH - 1);
			else
				--SerialQIn;
		
		// If we don't already know about the error, we'll clear the queue ok flag and set the flag
		// to send the error message.
			
		if (FlgSerialQueueOk)
		{
			FlgSerialQueueOk = 0;
			FlgSendOFErrMsg = 1;
		}
		
		// We've had a buffer overflow, so obviously there are messages in the queue, so we'll enable
		// the UART transmit interrupts as we exit
		
		TXIE = 1;
		return;		// EXIT
	}

	// Ok, we've saved all three bytes successfully into the Message Queue. Clear the error message flag,
	// and set the queue is ok flag if we've sent at least 138 messages successfully (2 Queue's worth)
	
	FlgSendOFErrMsg = 0;

	if (FlgSerialQueueOk == 0)
	{
		++SerialQNoOFCount;
		if (SerialQNoOFCount >= 138)
		{
			FlgSerialQueueOk = 1;
			SerialQNoOFCount = 0;
		}
	}	
		
	// Re-enable UART Transmit interrupts
		
	TXIE = 1;
}




/*-------------------------------------------------------------------------------------------------\
| Function:  	ProcessRcvdMsgs
|
| Purpose:   	Process a message that was received over the (serial bus), probably from the Set Box.
|
| Discussion:	Although the Logic Module mostly transmits messages, it must listen and execute
|					messages coming from the Set Box. In particular, there are three commands that must
|					be executed: 1) Set an AOA cutoff, 2) Set a Flaps Cutoff, 3) Reset all EEPROM values
|					to default values. Once these commands have been executed, an acknowledge message
|					must be sent back to the Set Box.
|
| Expects:		-
|					
|   Date   | Notes
|----------|---------------------------------------------------------------------------------------
| 01/01/00 | Revamped to match SB's function
| 12/11/00 | Created.
\--------------------------------------------------------------------------------------------------*/

void	ProcessRcvdMsg (void)
{

	static	bit	ResetMsgRcvd;
	
	// Check to see if the received message was a valid Set AOA Position message.
	
	//WAS LIKE: if ( (RcvdMsgID == 0x9F) && (RcvdMsgData > 0x00) && (RcvdMsgData <= 0x05) )


}


